/***************************************************************************
 *   Copyright (C) 2009-2011 by Conrad Hoffmann                            *
 *   ch@bitfehler.net                                                      *
 *   http://code.bitfehler.net                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2,       *
 *   as published by the Free Software Foundation,                         *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "serverstatusapplet.h"

#include <QAction>
#include <QGraphicsLayoutItem>
#include <QGraphicsGridLayout>
#include <QLabel>

#include <KAboutApplicationDialog>
#include <KAboutData>
#include <KConfigDialog>
#include <KDebug>
#include <KGlobal>
#include <KNotification>
#include <KNotifyConfigWidget>

#include <Plasma/Label>
#include <Plasma/ToolTipContent>
#include <Plasma/ToolTipManager>

#include "check.h"
#include "server.h"
#include "servericon.h"
#include "serverstatus.h"
#include "serverstatusconfig.h"


namespace ServerStatus
{

class AppletPrivate
{
public:
	
	// This is the main widget.
	QGraphicsWidget* widget;
	
	// A list of all widgets currently contained in the main widget.
	QList<QGraphicsItem*> widgets;

	// The config item for this applet instance.
	Config *config;
	
	// This is saved because it's needed for the notifications.
	KComponentData *compData;
	
	// The about dialog that can be shown via the context menu.
	KAboutApplicationDialog* about;
};

}

using namespace ServerStatus;

Applet::Applet(QObject *parent, const QVariantList &args) : Plasma::PopupApplet(parent, args)
{
	d = new AppletPrivate;
	
	// We need some custom AboutData to make notifications work correctly.
	KAboutData aboutData("plasma-applet-serverstatus", 0, ki18n("Server Status Widget"),
		"1.5", ki18n("Plasma applet to monitor remote hosts"), KAboutData::License_GPL,
		ki18n("Copyright (c) 2009-2011, Conrad Hoffmann"), KLocalizedString(), "http://code.bitfehler.net");
	aboutData.setProgramIconName("network-server");
	d->compData = new KComponentData(aboutData);
	
	// Create a nice "about" dialog
	d->about = new KAboutApplicationDialog(d->compData->aboutData());
	d->about->setWindowModality(Qt::NonModal);
	
	// This will get us the standard applet background, for free!
	setBackgroundHints(DefaultBackground);
	
	// Allow free resizing (i.e. must not be a square).
	setAspectRatioMode(Plasma::IgnoreAspectRatio);
	
	// Set a sane default size if the widget starts for the first time. If it's
	// not the first time, the size is set automatically before init().
	// Also, give it a layout.
	d->widget = new QGraphicsWidget(this);
	d->widget->setLayout(new QGraphicsGridLayout());
	d->widget->layout()->setContentsMargins(10, 10, 10, 10);
	d->widget->resize(250, 50);
}

Applet::~Applet()
{
	this->hide();
	delete d->compData;
	delete d->config;
	delete d->about;
	delete d;
}

void Applet::init()
{
	KConfigGroup cfg = config();
	d->config = new Config(this, d->compData->componentName());
	d->config->loadConfig(cfg);
	connect(d->config, SIGNAL(configChanged()), this, SLOT(slotConfigChanged()));

	rebuildLayout();

	foreach (Server* server, d->config->servers()) {
		server->startChecks();
	}
	
	// Set this as kind of a dummy tool tip, so that the slotToolTipAboutToShow will be triggered.
	// The actual tool tip is constructed in that slot.
	Plasma::ToolTipContent c(i18nc("Name of this applet", "Server Status"), 
							 i18n("No servers configured"), KIcon("network-server"));
	Plasma::ToolTipManager::self()->setContent(this, c);
}

QGraphicsWidget* Applet::graphicsWidget()
{
    return d->widget;
}

QList<QAction*> Applet::contextualActions()
{
	KIcon icon("network-server");
    QList<QAction*> result;
	QAction* aboutAction = new QAction(icon, i18n("About Server Status"), this);
	connect(aboutAction, SIGNAL(triggered()), d->about, SLOT(open()));
	result << aboutAction;
	return result;
}

void Applet::rebuildLayout()
{
	// Remove and delete all currently displayed items. There may be a more 
	// decent way to do this, but they have to be hidden _and_ removed.
	QGraphicsGridLayout* layout = dynamic_cast<QGraphicsGridLayout*>(d->widget->layout());

	for (int i = d->widgets.size() - 1; i >= 0; --i) {
		QGraphicsItem* w = dynamic_cast<QGraphicsItem*>(layout->itemAt(i));
		if (w != 0L) {
			layout->removeAt(i);
			w->hide();
			delete w;
		}
	}
	d->widgets.clear();
		
	QList<Server*> servers = d->config->servers();
	int size = d->config->minIconSize();

	if (!servers.empty()) {
		for (int i = 0; i < servers.size(); ++i) {
			Server* server = servers.at(i);
			
			// Create an icon for this server.
			ServerIcon* icon = new ServerIcon(d->config, d->widget);
			icon->setIcon(d->config->iconForStatus(server->check()->result()));
#if KDE_VERSION_MINOR > 4
			// This was added in KDE 4.5. The fallback method does not behave as nicely, but it's ok.
			icon->setMinimumIconSize(QSizeF(size, size));
#else
			icon->setMinimumSize(QSizeF(size, size));
#endif
			connect(server, SIGNAL(statusChanged(const ServerStatus::Status&)),
					icon, SLOT(slotChangeStatus(const ServerStatus::Status&)));
			disconnect(server, 0, this, 0);
			connect(server, SIGNAL(statusChanged(const ServerStatus::Server*, const ServerStatus::Status&)),
					this, SLOT(slotServerStatusChanged(const ServerStatus::Server*, const ServerStatus::Status&)));
			connect(server, SIGNAL(statusChanged(const ServerStatus::Status&)),
					this, SLOT(slotUpdateIcon()));
			layout->addItem(icon, i, 0);
			d->widgets.append(icon);

			// In some older versions of Qt, label->setFont did not work - thus also the call to set
			// the label of the native QLabel behind it. It may not be neccessary anymore, but it
			// can't harm either.
			Plasma::Label* label = new Plasma::Label;
			label->setFont(d->config->serverFont());
			label->nativeWidget()->setFont(d->config->serverFont());
			label->setText(server->name());
			
			layout->addItem(label, i, 1);
			d->widgets.append(label);
			
			d->widget->setMinimumSize(d->widget->layout()->minimumSize());
		}
	}
	else {
		ServerIcon* icon = new ServerIcon(d->config, d->widget);
		icon->setIcon("network-server");
#if KDE_VERSION_MINOR > 4
		// This was added in KDE 4.5. The fallback method does not behave as nicely, but it's ok.
		icon->setMinimumIconSize(QSizeF(size, size));
#else
		icon->setMinimumSize(QSizeF(size, size));
#endif
		layout->addItem(icon, 0, 0);
		d->widgets.append(icon);
		
		Plasma::Label* label = new Plasma::Label;
		label->setText(i18n("No servers configured"));
		layout->addItem(label, 0, 1);
		d->widgets.append(label);

		d->widget->setMinimumSize(d->widget->layout()->minimumSize());
	}
	
	d->widget->layout()->invalidate();
	update();
	
	slotUpdateIcon();
}

void Applet::slotUpdateIcon()
{
	// This will be set to the icon that should be used in case the applet is in a panel. 
	// Set a sane default, the standard icon.
	QString popupIcon("network-server");
	
	// We keep track of the worst status encountered and use that as icon, if applicable.
	int worstStatus = -1;

	QList<Server*> servers = d->config->servers();
	for (int i = 0; i < servers.size(); ++i) {
		Server* server = servers.at(i);
		
		// Check if this status is worse than what we've seen so far.
		const Status& status = server->check()->result();
		if (status.type() > worstStatus) {
			popupIcon = d->config->iconForStatus(status);
			worstStatus = status.type();
		}
	}
	setPopupIcon(popupIcon);
}

void Applet::createConfigurationInterface(KConfigDialog *parent)
{
	// Let the Config object handle this.
    d->config->createConfigurationInterface(parent);
}

void Applet::toolTipAboutToShow()
{
	// Generate a nice summary to show as tool tip.
	
	// If there are no servers, the default tool tip set during init is fine.
	QList<Server*> servers = d->config->servers();
	if (servers.empty()) {
		return;
	}

	// We keep track of the worst status encountered.
	Status& worstStatus = Status::None;
	
	// The servers which have the worst status. This is used to make a short list.
	QStringList worstServers;

	for (int i = 0; i < servers.size(); ++i) {
		Server* server = servers.at(i);
		
		// Check if this status is worse than what we've seen so far.
		const Status& status = server->check()->result();
		if (status.type() > worstStatus.type()) {
			// This is the worst status so far, so start a new list.
			worstServers.clear();
			worstServers << server->name();
			worstStatus = status;
		}
		else if (status.type() == worstStatus.type()) {
			worstServers << server->name();
		}
	}
	
	QString summary = i18ncp("Tooltip summary", "1 server with status <i>%2</i>", 
							 "%1 servers with status <i>%2</i>", worstServers.length(), 
							 worstStatus.name());
	// If all servers are up we don't need the names.
	QString serverNames = worstStatus.type() > Status::Up ? worstServers.join(", ") : QString();
	
	Plasma::ToolTipContent c(summary, serverNames, KIcon(d->config->iconForStatus(worstStatus)));
	Plasma::ToolTipManager::self()->setContent(this, c);
}

void Applet::slotConfigChanged()
{
	KConfigGroup cfg = config();
	d->config->saveConfig(cfg);
	rebuildLayout();
}

void Applet::slotServerStatusChanged(const Server* server, const Status& status)
{
	kDebug(0) << "Notify:" << status.eventName();
	KNotification *notification= new KNotification(status.eventName());
	notification->setComponentData(*d->compData);
#if KDE_VERSION_MINOR > 2
	// This was added in KDE 4.3
	notification->setTitle(i18nc("@info Notification title", "Server status notification"));
#endif
	QString text = i18nc("@info:status Notification text", 
						 "<b>New status for server <i>%1</i>: <i>%2</i></b>", 
						 server->name(), status.name());
	text.append("<br><br>" + status.description());
	notification->setText(text);
	notification->addContext("server", server->id());
	notification->sendEvent();
}


#include "serverstatusapplet.moc"
